#include <source_location>
#include <jlog/ansi_escape_codes.hpp>
#include <jlog/jlog.hpp>
#include <string>
#include <fstream>

namespace jlog {

    std::string get_timestamp()
    {
        using namespace std::chrono;
        auto const timestamp = current_zone()->to_local(system_clock::now());
        auto dp = floor<days>(timestamp);
        year_month_day ymd{floor<days>(timestamp)};
        hh_mm_ss time{floor<milliseconds>(timestamp-dp)};
        auto y = ymd.year();
        auto m = ymd.month();
        auto d = ymd.day();
        auto h = time.hours();
        auto M = time.minutes();
        auto s = time.seconds();
        auto ms = time.subseconds();
        return std::format("{}-{}-{} {}:{}:{}.{}", y, m, d, h.count(), M.count(), s.count(), ms.count());
    }

    void log_to_console(const std::string &message)
    {
        std::cout << message;
    }

    void log_to_file(const std::string &message)
    {
        std::ofstream latest_log("latest.log", std::ios_base::app);
        latest_log << message;
        latest_log.close();

    }

    void log(std::vector<token> tokens) {
        for (token t : tokens) {
            if (t.delimiter != "") {
                log_to_console(std::format("{}{}{}{}{} ", t.colorCode, t.delimiter[0], t.content, t.delimiter[1], ansi_escape_codes::RESET));
                log_to_file(std::format("{}{}{} ", t.delimiter[0], t.content, t.delimiter[1]));
            } else {
                log_to_console(std::format("{}{}{} ", t.colorCode, t.content, ansi_escape_codes::RESET));
                log_to_file(t.content + " ");
            }
        }
        log_to_console("\n");
        log_to_file("\n");
    }

    // Generic formatters for building loggers.
    std::vector<token> trace_format(
        const std::string& func,
        const std::string& file,
        int line)
    {
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        return {trace, filedata};
    }

    std::vector<token> log_format(
        const std::string& severity_name,
        const std::string& message,
        const std::string& severity_colorCode)
    {
        auto severity = token{.colorCode = severity_colorCode, .content = severity_name};
        auto content = token{.content = message, .delimiter = ""};
        return {severity, content};
    }

    std::vector<token> log_detailed_format(
    const std::string& severity_name,
    const std::string& message,
    const std::string& func,
    const std::string& file,
    int line,
    const std::string& severity_colorCode)
    {
        std::vector<token> tokens;
        auto timestamp = token{.content = jlog::get_timestamp()};
        auto tf_tokens = trace_format(func, file, line);
        auto lf_tokens = log_format(severity_name, message, severity_colorCode);
        tokens.push_back(timestamp);
        tokens.insert(tokens.end(), tf_tokens.begin(), tf_tokens.end());
        tokens.insert(tokens.end(), lf_tokens.begin(), lf_tokens.end());
        return tokens;
    }



    // Predefined generic loggers for jlog.
    std::vector<token> info_format(const std::string& message)
    {
        return log_format("INFO", message);
    }

    std::vector<token> info_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line)
    {
        return log_detailed_format("INFO", message, func, file, line);
    }

    std::vector<token> warning_format(const std::string& message)
    {
        return log_format("INFO", message, ansi_escape_codes::FG_YELLOW);
    }

    std::vector<token> warning_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line)
    {
        return log_detailed_format("WARNING", message, func, file, line, ansi_escape_codes::FG_YELLOW);
    }

    std::vector<token> error_format(const std::string& message)
    {
        return log_format("ERROR", message, ansi_escape_codes::FG_RED);
    }

    std::vector<token> error_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line)
    {
        return log_detailed_format("ERROR", message, func, file, line, ansi_escape_codes::FG_RED);
    }

    std::vector<token> fatal_format(const std::string& message)
    {
        return log_format("FATAL", message, ansi_escape_codes::FG_BRIGHT_RED);
    }

    std::vector<token> fatal_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line)
    {
        return log_detailed_format("FATAL", message, func, file, line, ansi_escape_codes::FG_BRIGHT_RED);
    }

    std::vector<token> verbose_format(const std::string& message)
    {
        return log_format("VERBOSE", message, ansi_escape_codes::FG_CYAN);
    }

    std::vector<token> verbose_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line)
    {
        return log_detailed_format("VERBOSE", message, func, file, line, ansi_escape_codes::FG_CYAN);
    }

    std::vector<token> debug_format(const std::string& message)
    {
        return log_format("DEBUG", message, ansi_escape_codes::FG_GREEN);
    }

    std::vector<token> debug_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line)
    {
        return log_detailed_format("DEBUG", message, func, file, line, ansi_escape_codes::FG_GREEN);
    }
}
